@@ -1,8 +1,9 @@
-from __future__ import division, print_function, absolute_import
+from __future__ import absolute_import, division, print_function
 
 from . import boreholes
 from . import gfunction
 from . import heat_transfer
 from . import load_aggregation
+from . import networks
 from . import pipes
 from . import utilities
