@@ -1,15 +1,14 @@
-from __future__ import division, print_function, absolute_import
+from __future__ import absolute_import, division, print_function
 
-from functools import partial
-import numpy as np
-from multiprocessing import Pool
-from scipy.interpolate import interp1d as interp1d
-from scipy.constants import pi
 import time as tim
 
+import numpy as np
+from scipy.constants import pi
+from scipy.interpolate import interp1d as interp1d
+
 from .boreholes import Borehole
-from .heat_transfer import finite_line_source as FLS
-from .heat_transfer import similarities
+from .heat_transfer import thermal_response_factors
+from .networks import network_thermal_resistance
 
 
 def uniform_heat_extraction(boreholes, time, alpha, use_similarities=True,
@@ -38,7 +37,7 @@ def uniform_heat_extraction(boreholes, time, alpha, use_similarities=True,
         difference between the two distances (abs(d1-d2)) is below tolerance.
         Default is 0.1.
     tol : float, optional
-        Relative tolerance on length and depth. Two lenths H1, H2
+        Relative tolerance on length and depth. Two lengths H1, H2
         (or depths D1, D2) are considered equal if abs(H1 - H2)/H2 < tol.
         Default is 1.0e-6.
     processes : int, optional
@@ -146,7 +145,7 @@ def uniform_temperature(boreholes, time, alpha, nSegments=12, method='linear',
         difference between the two distances (abs(d1-d2)) is below tolerance.
         Default is 0.1.
     tol : float, optional
-        Relative tolerance on length and depth. Two lenths H1, H2
+        Relative tolerance on length and depth. Two lengths H1, H2
         (or depths D1, D2) are considered equal if abs(H1 - H2)/H2 < tol.
         Default is 1.0e-6.
     processes : int, optional
@@ -292,8 +291,8 @@ def equal_inlet_temperature(boreholes, UTubes, m_flow, cp, time, alpha,
         Model for pipes inside each borehole.
     m_flow : float or array
         Fluid mass flow rate per borehole (in kg/s).
-    cp : fluid specific isobaric heat capacity (in J/kg.K)
-        Model with fluid properties.
+    cp : float
+        Fluid specific isobaric heat capacity (in J/kg.K).
     time : float or array
         Values of time (in seconds) for which the g-function is evaluated.
     alpha : float
@@ -314,7 +313,7 @@ def equal_inlet_temperature(boreholes, UTubes, m_flow, cp, time, alpha,
         difference between the two distances (abs(d1-d2)) is below tolerance.
         Default is 0.1.
     tol : float, optional
-        Relative tolerance on length and depth. Two lenths H1, H2
+        Relative tolerance on length and depth. Two lengths H1, H2
         (or depths D1, D2) are considered equal if abs(H1 - H2)/H2 < tol.
         Default is 1.0e-6.
     processes : int, optional
@@ -467,39 +466,51 @@ def equal_inlet_temperature(boreholes, UTubes, m_flow, cp, time, alpha,
     return gFunction
 
 
-def thermal_response_factors(
-        boreSegments, time, alpha, use_similarities=True,
-        splitRealAndImage=True, disTol=0.1, tol=1.0e-6, processes=None,
-        disp=False):
+def mixed_inlet_temperature(network, m_flow, cp,
+                            time, alpha, method='linear', nSegments=12,
+                            use_similarities=True, disTol=0.1, tol=1.0e-6,
+                            processes=None, disp=False):
     """
-    Evaluate segment-to-segment thermal response factors.
+    Evaluate the g-function with mixed inlet fluid temperatures.
 
-    This function goes through the list of borehole segments and evaluates
-    the segments-to-segment response factors for all times in time.
+    This function superimposes the finite line source (FLS) solution to
+    estimate the g-function of a geothermal bore field. Each borehole is
+    modeled as a series of finite line source segments, as proposed in
+    [#Cimmino2018]_. The piping configurations between boreholes can be any
+    combination of series and parallel connections.
 
     Parameters
     ----------
-    boreSegments : list of Borehole objects
-        List of borehole segments.
+    network : Network objects
+        List of boreholes included in the bore field.
+    m_flow : float or array
+        Total mass flow rate into the network or inlet mass flow rates
+        into each circuit of the network (in kg/s). If a float is supplied,
+        the total mass flow rate is split equally into all circuits.
+    cp : float or array
+        Fluid specific isobaric heat capacity (in J/kg.degC).
+        Must be the same for all circuits (a single float can be supplied).
     time : float or array
         Values of time (in seconds) for which the g-function is evaluated.
     alpha : float
         Soil thermal diffusivity (in m2/s).
+    nSegments : int, optional
+        Number of line segments used per borehole.
+        Default is 12.
+    method : string, optional
+        Interpolation method used for segment-to-segment thermal response
+        factors. See documentation for scipy.interpolate.interp1d.
+        Default is 'linear'.
     use_similarities : bool, optional
         True if similarities are used to limit the number of FLS evaluations.
         Default is True.
-    splitRealAndImage : bool, optional
-        Set to True if similarities are evaluated separately for real and image
-        sources. Set to False if similarities are evaluated for the sum of the
-        real and image sources.
-        Default is True.
     disTol : float, optional
         Absolute tolerance (in meters) on radial distance. Two distances
         (d1, d2) between two pairs of boreholes are considered equal if the
         difference between the two distances (abs(d1-d2)) is below tolerance.
         Default is 0.1.
     tol : float, optional
-        Relative tolerance on length and depth. Two lenths H1, H2
+        Relative tolerance on length and depth. Two lengths H1, H2
         (or depths D1, D2) are considered equal if abs(H1 - H2)/H2 < tol.
         Default is 1.0e-6.
     processes : int, optional
@@ -512,117 +523,145 @@ def thermal_response_factors(
 
     Returns
     -------
-    h_ij : array
-        Segment-to-segment thermal response factors.
+    gFunction : float or array
+        Values of the g-function
+
+    Examples
+    --------
+    >>> b1 = gt.boreholes.Borehole(H=150., D=4., r_b=0.075, x=0., y=0.)
+    >>> b2 = gt.boreholes.Borehole(H=150., D=4., r_b=0.075, x=5., y=0.)
+    >>> Utube1 = gt.pipes.SingleUTube(pos=[(-0.05, 0), (0, -0.05)],
+                                      r_in=0.015, r_out=0.02,
+                                      borehole=b1,k_s=2, k_g=1, R_fp=0.1)
+    >>> Utube2 = gt.pipes.SingleUTube(pos=[(-0.05, 0), (0, -0.05)],
+                                      r_in=0.015, r_out=0.02,
+                                      borehole=b1,k_s=2, k_g=1, R_fp=0.1)
+    >>> bore_connectivity = [-1, 0]
+    >>> network = gt.networks.Network([b1, b2], [Utube1, Utube2], bore_connectivity)
+    >>> time = np.array([1.0*10**i for i in range(4, 12)])
+    >>> m_flow = 0.25
+    >>> cp = 4000.
+    >>> alpha = 1.0e-6
+    >>> gt.gfunction.mixed_inlet_temperature(network, m_flow, cp, time, alpha)
+    array([0.63782415, 1.63304116, 2.72191316, 4.04091713, 5.98240458,
+       7.77216202, 8.66195828, 8.77567215])
+
+    References
+    ----------
+    .. [#Cimmino2018] Cimmino, M. (2018). g-Functions for bore fields with
+       mixed parallel and series connections considering the axial fluid
+       temperature variations. Proceedings of the IGSHPA Sweden Research Track
+       2018. Stockholm, Sweden. pp. 262-270.
 
     """
-    # Total number of line sources
-    nSources = len(boreSegments)
-    # Number of time values
-    nt = len(np.atleast_1d(time))
-    # Prepare pool of workers for parallel computation
-    pool = Pool(processes=processes)
+    if disp:
+        print(60*'-')
+        print('Calculating g-function for mixed inlet fluid temperatures')
+        print(60*'-')
     # Initialize chrono
     tic = tim.time()
+    # Number of boreholes
+    nBoreholes = network.nBoreholes
+    # Total number of line sources
+    nSources = nSegments*nBoreholes
+    # Number of time values
+    nt = len(np.atleast_1d(time))
+    # Initialize g-function
+    gFunction = np.zeros_like(np.atleast_1d(time))
+    # Initialize segment heat extraction rates
+    Q = np.zeros((nSources, nt))
 
-    # Initialize segment-to-segment response factors
-    h_ij = np.zeros((nSources, nSources, nt))
-    # Calculation is based on the choice of use_similarities
-    if use_similarities:
-        # Calculations with similarities
-        if disp:
-            print('Identifying similarities ...')
-        (nSimPos, simPos, disSimPos, HSimPos, DSimPos,
-         nSimNeg, simNeg, disSimNeg, HSimNeg, DSimNeg) = \
-            similarities(boreSegments,
-                         splitRealAndImage=splitRealAndImage,
-                         disTol=disTol,
-                         tol=tol,
-                         processes=processes)
+    # Ground thermal conductivity
+    k_s = network.p[0].k_s
+    # Split boreholes into segments
+    boreSegments = _borehole_segments(network.b, nSegments)
+    # Vector of time values
+    t = np.atleast_1d(time).flatten()
+    # Calculate segment to segment thermal response factors
+    h_ij = thermal_response_factors(
+        boreSegments, t, alpha, use_similarities=use_similarities,
+        splitRealAndImage=True, disTol=disTol, tol=tol, processes=processes,
+        disp=disp)
+    toc1 = tim.time()
 
-        toc1 = tim.time()
-        if disp:
-            print('{} sec'.format(toc1 - tic))
-            print('Calculating segment to segment response factors ...')
-
-        # Initialize FLS solution for the real source
-        hPos = np.zeros(nt)
-        # Initialize FLS solution for the image source
-        hNeg = np.zeros(nt)
-
-        # Similarities for real sources
-        for s in range(nSimPos):
-            n1 = simPos[s][0][0]
-            n2 = simPos[s][0][1]
-            b1 = boreSegments[n1]
-            b2 = boreSegments[n2]
-            if splitRealAndImage:
-                # FLS solution for real source only
-                func = partial(FLS, alpha=alpha, borehole1=b1, borehole2=b2,
-                               reaSource=True, imgSource=False)
-            else:
-                # FLS solution for combined real and image sources
-                func = partial(FLS, alpha=alpha, borehole1=b1, borehole2=b2,
-                               reaSource=True, imgSource=True)
-            # Evaluate the FLS solution at all times in parallel
-            hPos = np.array(pool.map(func, np.atleast_1d(time)))
-            # Assign thermal response factors to similar segment pairs
-            for (i, j) in simPos[s]:
-                h_ij[j, i, :] = hPos
-                h_ij[i, j, :] = b2.H/b1.H * hPos
-
-        # Similarities for image sources (only if splitRealAndImage=True)
-        if splitRealAndImage:
-            for s in range(nSimNeg):
-                n1 = simNeg[s][0][0]
-                n2 = simNeg[s][0][1]
-                b1 = boreSegments[n1]
-                b2 = boreSegments[n2]
-                # FLS solution for image source only
-                func = partial(FLS, alpha=alpha, borehole1=b1, borehole2=b2,
-                               reaSource=False, imgSource=True)
-                # Evaluate the FLS solution at all times in parallel
-                hNeg = np.array(pool.map(func, time))
-                # Assign thermal response factors to similar segment pairs
-                for (i, j) in simNeg[s]:
-                    h_ij[j, i, :] = h_ij[j, i, :] + hNeg
-                    h_ij[i, j, :] = b2.H/b1.H * h_ij[j, i, :]
+    if disp:
+        print('Building and solving system of equations ...')
+    # -------------------------------------------------------------------------
+    # Build a system of equation [A]*[X] = [B] for the evaluation of the
+    # g-function. [A] is a coefficient matrix, [X] = [Qb,Tb,Tf_in] is a state
+    # space vector of the borehole heat extraction rates, borehole wall
+    # temperatures and inlet fluid temperature (into the bore field),
+    # [B] is a coefficient vector.
+    # -------------------------------------------------------------------------
 
+    # Segment lengths
+    Hb = np.array([[b.H for b in boreSegments]])
+    # Vector of time steps
+    dt = np.hstack((t[0], t[1:] - t[:-1]))
+    if not np.isscalar(time) and len(time) > 1:
+        # Spline object for thermal response factors
+        h_dt = interp1d(np.hstack((0., t)),
+                        np.dstack((np.zeros((nSources,nSources)), h_ij)),
+                        kind=method, axis=2)
+        # Thermal response factors evaluated at t=dt
+        h_dt = h_dt(dt)
     else:
-        # Calculations without similarities
-        if disp:
-            print('Calculating segment to segment response factors ...')
-        for i in range(nSources):
-            # Segment to same-segment thermal response factor
-            # FLS solution for combined real and image sources
-            b2 = boreSegments[i]
-            func = partial(FLS, alpha=alpha, borehole1=b2, borehole2=b2)
-            # Evaluate the FLS solution at all times in parallel
-            h = np.array(pool.map(func, time))
-            h_ij[i, i, :] = h
+        h_dt = h_ij
+    # Thermal response factor increments
+    dh_ij = np.concatenate((h_ij[:,:,0:1], h_ij[:,:,1:]-h_ij[:,:,:-1]), axis=2)
 
-            # Segment to other segments thermal response factor
-            for j in range(i+1, nSources):
-                b1 = boreSegments[j]
-                # Evaluate the FLS solution at all times in parallel
-                func = partial(FLS, alpha=alpha, borehole1=b1, borehole2=b2)
-                h = np.array(pool.map(func, time))
-                h_ij[i, j, :] = h
-                h_ij[j, i, :] = b2.H / b1.H * h_ij[i, j, :]
+    # Energy balance on borehole segments:
+    # [Q_{b,i}] = [a_in]*[T_{f,in}] + [a_{b,i}]*[T_{b,i}]
+    a_in, a_b = network.coefficients_borehole_heat_extraction_rate(
+            m_flow, cp, nSegments)
+    A_eq2 = np.hstack((np.eye(nSources), a_b/(2.0*pi*k_s*Hb.T), a_in/(2.0*pi*k_s*Hb.T)))
+    B_eq2 = np.zeros(nSources)
+
+    # Energy conservation: sum([Qb*Hb]) = sum([Hb])
+    A_eq3 = np.hstack((Hb, np.zeros((1, nSources + 1))))
+    B_eq3 = np.atleast_1d(np.sum(Hb))
+
+    # Build and solve the system of equations at all times
+    for p in range(nt):
+        # Current thermal response factor matrix
+        h_ij_dt = h_dt[:,:,p]
+        # Reconstructed load history
+        Q_reconstructed = load_history_reconstruction(t[0:p+1], Q[:,0:p+1])
+        # Borehole wall temperature for zero heat extraction at current step
+        Tb_0 = _temporal_superposition(dh_ij, Q_reconstructed)
+        # Spatial superposition: [Tb] = [Tb0] + [h_ij_dt]*[Qb]
+        A_eq1 = np.hstack((h_ij_dt,
+                           -np.eye(nSources),
+                           np.zeros((nSources, 1))))
+        B_eq1 = -Tb_0
+        # Assemble equations
+        B = np.hstack((B_eq1, B_eq2, B_eq3))
+        A = np.vstack((A_eq1, A_eq2, A_eq3))
+        # Solve the system of equations
+        X = np.linalg.solve(A, B)
+        # Store calculated heat extraction rates
+        Q[:,p] = X[0:nSources]
+        # The gFunction is equal to the average borehole wall temperature
+        Tf_in = X[-1]
+        Tf_out = Tf_in - 2*pi*k_s*np.sum(Hb)/(m_flow*cp)
+        Tf = 0.5*(Tf_in + Tf_out)
+        Rfield = network_thermal_resistance(network, m_flow, cp)
+        Tb_eff = Tf - 2*pi*k_s*Rfield
+        gFunction[p] = Tb_eff
 
     toc2 = tim.time()
+
     if disp:
-        print('{} sec'.format(toc2 - tic))
+        print('{} sec'.format(toc2 - toc1))
+        print('Total time for g-function evaluation: {} sec'.format(
+                toc2 - tic))
+        print(60*'-')
 
-    # Close pool of workers
-    pool.close()
-    pool.join()
-
-    # Return 2d array if time is a scalar
+    # Return float if time is a scalar
     if np.isscalar(time):
-        h_ij = h_ij[:,:,0]
+        gFunction = np.asscalar(gFunction)
 
-    return h_ij
+    return gFunction
 
 
 def load_history_reconstruction(time, Q):
@@ -692,7 +731,7 @@ def _borehole_segments(boreholes, nSegments):
         for i in range(nSegments):
             # Divide borehole into segments of equal length
             H = b.H / nSegments
-            # Burried depth of the i-th segment
+            # Buried depth of the i-th segment
             D = b.D + i * b.H / nSegments
             # Add to list of segments
             boreSegments.append(Borehole(H, D, b.r_b, b.x, b.y))
@@ -708,7 +747,7 @@ def _temporal_superposition(dh_ij, Q):
     dh_ij : array
         Values of the segment-to-segment thermal response factor increments at
         the given time step.
-    Q_reconstructed : array
+    Q : array
         Heat extraction rates of all segments at all times.
 
     Returns
