@@ -1,4 +1,4 @@
-from __future__ import division, print_function, absolute_import
+from __future__ import absolute_import, division, print_function
 
 import numpy as np
 from scipy.constants import pi
@@ -24,7 +24,7 @@ class _BasePipe(object):
 
     """
     def __init__(self, borehole):
-        self.borehole = borehole
+        self.b = borehole
         self.nPipes = 1
         self.nInlets = 1
         self.nOutlets = 1
@@ -530,6 +530,69 @@ class _BasePipe(object):
 
         return a_in, a_b
 
+    def visualize_pipes(self):
+        """
+        Plot the cross-section view of the borehole.
+
+        Returns
+        -------
+        fig : figure
+            Figure object (matplotlib).
+
+        """
+        import matplotlib.pyplot as plt
+        from matplotlib.ticker import AutoMinorLocator
+
+        # Initialize figure
+        LW = 1.5    # Line width
+        FS = 12.    # Font size
+
+        plt.rc('figure', figsize=(80.0/25.4, 80.0/25.4))
+        fig = plt.figure()
+        ax = fig.add_subplot(111)
+
+        # Borehole wall outline
+        borewall = plt.Circle((0., 0.), radius=self.b.r_b,
+                              fill=False, linestyle='--', linewidth=LW)
+        ax.add_patch(borewall)
+
+        # Pipes
+        for i in range(self.nPipes):
+            # Coordinates of pipes
+            (x_in, y_in) = self.pos[i]
+            (x_out, y_out) = self.pos[i + self.nPipes]
+
+            # Pipe outline (inlet)
+            pipe_in_in = plt.Circle((x_in, y_in), radius=self.r_in,
+                                    fill=False, linestyle='-', linewidth=LW)
+            pipe_in_out = plt.Circle((x_in, y_in), radius=self.r_out,
+                                     fill=False, linestyle='-', linewidth=LW)
+            ax.text(x_in, y_in, i + 1,
+                    ha="center", va="center", size=FS)
+
+            # Pipe outline (outlet)
+            pipe_out_in = plt.Circle((x_out, y_out), radius=self.r_in,
+                                     fill=False, linestyle='-', linewidth=LW)
+            pipe_out_out = plt.Circle((x_out, y_out), radius=self.r_out,
+                                      fill=False, linestyle='-', linewidth=LW)
+            ax.text(x_out, y_out, i + self.nPipes + 1,
+                    ha="center", va="center", size=FS)
+
+            ax.add_patch(pipe_in_in)
+            ax.add_patch(pipe_in_out)
+            ax.add_patch(pipe_out_in)
+            ax.add_patch(pipe_out_out)
+
+        # Configure figure axes
+        ax.set_xlabel('x (m)')
+        ax.set_ylabel('y (m)')
+        plt.axis('equal')
+        ax.xaxis.set_minor_locator(AutoMinorLocator())
+        ax.yaxis.set_minor_locator(AutoMinorLocator())
+        plt.tight_layout()
+
+        return fig
+
     def _initialize_stored_coefficients(self):
         nMethods = 8    # Number of class methods
         self._stored_coefficients = [() for i in range(nMethods)]
@@ -539,7 +602,6 @@ class _BasePipe(object):
         self._m_flow_cp_model_variables = np.empty(self.nInlets)
         self._nSegments_model_variables = np.nan
 
-        return
 
     def _set_stored_coefficients(self, m_flow, cp, nSegments, coefficients,
                                  method_id):
@@ -547,8 +609,6 @@ class _BasePipe(object):
         self._stored_m_flow_cp[method_id] = m_flow*cp
         self._stored_nSegments[method_id] = nSegments
 
-        return
-
     def _get_stored_coefficients(self, method_id):
         coefficients = self._stored_coefficients[method_id]
 
@@ -644,7 +704,7 @@ class SingleUTube(_BasePipe):
     r_in : float
         Inner radius (in meters) of the U-Tube pipes.
     r_out : float
-        Outter radius (in meters) of the U-Tube pipes.
+        Outer radius (in meters) of the U-Tube pipes.
     borehole : Borehole object
         Borehole class object of the borehole containing the U-Tube.
     k_s : float
@@ -652,7 +712,7 @@ class SingleUTube(_BasePipe):
     k_g : float
         Grout thermal conductivity (in W/m-K).
     R_fp : float
-        Fluid to outter pipe wall thermal resistance (m-K/W).
+        Fluid to outer pipe wall thermal resistance (m-K/W).
     J : int, optional
         Number of multipoles per pipe to evaluate the thermal resistances.
         Default is 2.
@@ -798,6 +858,8 @@ class SingleUTube(_BasePipe):
 
         Parameters
         ----------
+        z : float or array
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         m_flow : float or array
             Inlet mass flow rate (in kg/s).
         cp : float or array
@@ -834,7 +896,17 @@ class SingleUTube(_BasePipe):
     def _update_model_variables(self, m_flow, cp, nSegments):
         """
         Evaluate dimensionless resistances for Hellstrom (1991) solution.
+
+        Parameters
+        ----------
+        m_flow : float or array
+            Inlet mass flow rate (in kg/s).
+        cp : float or array
+            Fluid specific isobaric heat capacity (in J/kg.degC).
+        nSegments : int
+            Number of borehole segments.
         """
+
         # Format mass flow rate and heat capacity inputs
         self._format_inputs(m_flow, cp, nSegments)
         m_flow_in = self._m_flow_in
@@ -854,7 +926,17 @@ class SingleUTube(_BasePipe):
     def _format_inputs(self, m_flow, cp, nSegments):
         """
         Format mass flow rate and heat capacity inputs.
+
+        Parameters
+        ----------
+        m_flow : float or array
+            Inlet mass flow rate (in kg/s).
+        cp : float or array
+            Fluid specific isobaric heat capacity (in J/kg.degC).
+        nSegments : int
+            Number of borehole segments.
         """
+
         # Format mass flow rate inputs
         if np.isscalar(m_flow):
             # Mass flow rate in each fluid circuit
@@ -882,7 +964,13 @@ class SingleUTube(_BasePipe):
     def _f1(self, z):
         """
         Calculate function f1 from Hellstrom (1991)
+
+        Parameters
+        ----------
+        z : float
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         """
+
         f1 = np.exp(self._beta*z)*(np.cosh(self._gamma*z)
                                    - self._delta*np.sinh(self._gamma*z))
         return f1
@@ -890,7 +978,13 @@ class SingleUTube(_BasePipe):
     def _f2(self, z):
         """
         Calculate function f2 from Hellstrom (1991)
+
+        Parameters
+        ----------
+        z : float
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         """
+
         f2 = np.exp(self._beta*z)*self._beta12/self._gamma \
             * np.sinh(self._gamma*z)
         return f2
@@ -898,7 +992,13 @@ class SingleUTube(_BasePipe):
     def _f3(self, z):
         """
         Calculate function f3 from Hellstrom (1991)
+
+        Parameters
+        ----------
+        z : float
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         """
+
         f3 = np.exp(self._beta*z)*(np.cosh(self._gamma*z)
                                    + self._delta*np.sinh(self._gamma*z))
         return f3
@@ -906,7 +1006,13 @@ class SingleUTube(_BasePipe):
     def _f4(self, z):
         """
         Calculate function f4 from Hellstrom (1991)
+
+        Parameters
+        ----------
+        z : float
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         """
+
         A = self._delta*self._beta1 + self._beta2*self._beta12/self._gamma
         f4 = np.exp(self._beta*z) \
             * (self._beta1*np.cosh(self._gamma*z) - A*np.sinh(self._gamma*z))
@@ -915,7 +1021,13 @@ class SingleUTube(_BasePipe):
     def _f5(self, z):
         """
         Calculate function f5 from Hellstrom (1991)
+
+        Parameters
+        ----------
+        z : float
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         """
+
         B = self._delta*self._beta2 + self._beta1*self._beta12/self._gamma
         f5 = np.exp(self._beta*z) \
             * (self._beta2*np.cosh(self._gamma*z) + B*np.sinh(self._gamma*z))
@@ -924,7 +1036,13 @@ class SingleUTube(_BasePipe):
     def _F4(self, z):
         """
         Calculate integral of function f4 from Hellstrom (1991)
+
+        Parameters
+        ----------
+        z : float
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         """
+
         A = self._delta*self._beta1 + self._beta2*self._beta12/self._gamma
         C = self._beta1*self._beta + A*self._gamma
         S = - (self._beta1*self._gamma + self._beta*A)
@@ -936,7 +1054,13 @@ class SingleUTube(_BasePipe):
     def _F5(self, z):
         """
         Calculate integral of function f5 from Hellstrom (1991)
+
+        Parameters
+        ----------
+        z : float
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         """
+
         B = self._delta*self._beta2 + self._beta1*self._beta12/self._gamma
         C = self._beta2*self._beta - B*self._gamma
         S = - (self._beta2*self._gamma - self._beta*B)
@@ -962,7 +1086,7 @@ class MultipleUTube(_BasePipe):
     r_in : float
         Inner radius (in meters) of the U-Tube pipes.
     r_out : float
-        Outter radius (in meters) of the U-Tube pipes.
+        Outer radius (in meters) of the U-Tube pipes.
     borehole : Borehole object
         Borehole class object of the borehole containing the U-Tube.
     k_s : float
@@ -970,7 +1094,7 @@ class MultipleUTube(_BasePipe):
     k_g : float
         Grout thermal conductivity (in W/m-K).
     R_fp : float
-        Fluid to outter pipe wall thermal resistance (m-K/W).
+        Fluid to outer pipe wall thermal resistance (m-K/W).
     J : int, optional
         Number of multipoles per pipe to evaluate the thermal resistances.
         Default is 2.
@@ -1091,6 +1215,8 @@ class MultipleUTube(_BasePipe):
             a_in = np.linalg.multi_dot([e_u, d_u_m1, d_in])
             a_out = np.array([[1.0]])
             a_b = np.linalg.multi_dot([e_u, d_u_m1, d_b])
+        else:
+            raise NotImplementedError("Configuration '{}' not implemented.".format(self.config))
 
         return a_in, a_out, a_b
 
@@ -1166,6 +1292,8 @@ class MultipleUTube(_BasePipe):
             a_out = np.zeros((2*self.nPipes, self.nOutlets))
             a_b = np.linalg.multi_dot([e_d, c_u, d_u_m1, d_b]) \
                 + np.linalg.multi_dot([e_u, d_u_m1, d_b])
+        else:
+            raise NotImplementedError("Configuration '{}' not implemented.".format(self.config))
 
         return a_in, a_out, a_b
 
@@ -1253,6 +1381,8 @@ class MultipleUTube(_BasePipe):
 
         Parameters
         ----------
+        z : float or array
+            Depth (in meters) to evaluate the fluid temperature coefficients.
         m_flow : float or array
             Inlet mass flow rate (in kg/s).
         cp : float or array
@@ -1302,7 +1432,17 @@ class MultipleUTube(_BasePipe):
         """
         Evaluate eigenvalues and eigenvectors for the system of differential
         equations.
+
+        Parameters
+        ----------
+        m_flow : float or array
+            Inlet mass flow rate (in kg/s).
+        cp : float or array
+            Fluid specific isobaric heat capacity (in J/kg.degC).
+        nSegments : int
+            Number of borehole segments.
         """
+
         nPipes = self.nPipes
         # Format mass flow rate and heat capacity inputs
         self._format_inputs(m_flow, cp, nSegments)
@@ -1327,7 +1467,17 @@ class MultipleUTube(_BasePipe):
     def _format_inputs(self, m_flow, cp, nSegments):
         """
         Format mass flow rate and heat capacity inputs.
+
+        Parameters
+        ----------
+        m_flow : float or array
+            Inlet mass flow rate (in kg/s).
+        cp : float or array
+            Fluid specific isobaric heat capacity (in J/kg.degC).
+        nSegments : int
+            Number of borehole segments.
         """
+
         nPipes = self.nPipes
         # Format mass flow rate inputs
         # Mass flow rate in pipes
@@ -1365,7 +1515,7 @@ class IndependentMultipleUTube(MultipleUTube):
     r_in : float
         Inner radius (in meters) of the U-Tube pipes.
     r_out : float
-        Outter radius (in meters) of the U-Tube pipes.
+        Outer radius (in meters) of the U-Tube pipes.
     borehole : Borehole object
         Borehole class object of the borehole containing the U-Tube.
     k_s : float
@@ -1373,7 +1523,7 @@ class IndependentMultipleUTube(MultipleUTube):
     k_g : float
         Grout thermal conductivity (in W/m-K).
     R_fp : float
-        Fluid to outter pipe wall thermal resistance (m-K/W).
+        Fluid to outer pipe wall thermal resistance (m-K/W).
     J : int, optional
         Number of multipoles per pipe to evaluate the thermal resistances.
         Default is 2.
@@ -1633,6 +1783,39 @@ def thermal_resistances(pos, r_out, r_b, k_s, k_g, Rfp, J=2):
     return R, Rd
 
 
+def borehole_thermal_resistance(pipe, m_flow, cp):
+    """
+    Evaluate the effective borehole thermal resistance.
+
+    Parameters
+    ----------
+    pipe : pipe object
+        Model for pipes inside the borehole.
+    m_flow : float
+        Fluid mass flow rate (in kg/s).
+    cp : float
+        Fluid specific isobaric heat capacity (in J/kg.K)
+
+    Returns
+    -------
+    Rb : float
+        Effective borehole thermal resistance (m.K/W).
+
+    """
+    # Coefficient for T_{f,out} = a_out*T_{f,in} + [b_out]*[T_b]
+    a_out = np.asscalar(
+            pipe.coefficients_outlet_temperature(m_flow, cp, nSegments=1)[0])
+    # Coefficient for Q_b = [a_Q]*T{f,in} + [b_Q]*[T_b]
+    a_Q = np.asscalar(pipe.coefficients_borehole_heat_extraction_rate(
+            m_flow, cp, nSegments=1)[0])
+    # Borehole length
+    H = pipe.b.H
+    # Effective borehole thermal resistance
+    Rb = -0.5*H*(1. + a_out)/a_Q
+
+    return Rb
+
+
 def fluid_friction_factor_circular_pipe(m_flow, r_in, visc, den, epsilon,
                                         tol=1.0e-6):
     """
@@ -1735,9 +1918,12 @@ def convective_heat_transfer_coefficient_circular_pipe(m_flow, r_in, visc, den,
     # Darcy friction factor
     fDarcy = fluid_friction_factor_circular_pipe(m_flow, r_in, visc, den,
                                                  epsilon)
-    # Nusselt number from Gnielinski
-    Nu = 0.125*fDarcy * (Re - 1.0e3) * Pr / \
-        (1.0 + 12.7 * np.sqrt(0.125*fDarcy) * (Pr**(2.0/3.0) - 1.0))
+    if Re > 2300.:
+        # Nusselt number from Gnielinski
+        Nu = 0.125*fDarcy * (Re - 1.0e3) * Pr / \
+            (1.0 + 12.7 * np.sqrt(0.125*fDarcy) * (Pr**(2.0/3.0) - 1.0))
+    else:
+        Nu = 3.66
     h_fluid = k * Nu / D
 
     return h_fluid
@@ -1800,7 +1986,7 @@ def multipole(pos, r_p, r_b, k_s, k_g, Rfp, T_b, Q_p, J,
         line source approximation.
     Q_p : array
         Thermal energy flows (in W/m) from pipes.
-    T_b_av : float
+    T_b : float
         Average borehole wall temperature (in degC).
     eps : float, optional
         Iteration relative accuracy.
@@ -1878,6 +2064,8 @@ def multipole(pos, r_p, r_b, k_s, k_g, Rfp, T_b, Q_p, J,
             diff = np.max(np.abs(P_new-P)) - np.min(np.abs(P_new-P))
             eps_max = diff / diff0
             P = P_new
+    else:
+        P = np.zeros((n_p, 0))
 
     # --------------------------
     # Fluid temperatures(EQ. 32)
@@ -1939,7 +2127,37 @@ def multipole(pos, r_p, r_b, k_s, k_g, Rfp, T_b, Q_p, J,
 
 def _F_mk(Q_p, P, n_p, J, r_b, r_p, z, pikg, sigma):
     """
-    Complex matrix F_mk from Claesson and Hellstrom (2011), EQ. 34
+    Complex matrix F_mk from Claesson and Hellstrom (2011), EQ. 34.
+
+    Parameters
+    ----------
+    Q_p : array
+        Thermal energy flows (in W/m) from pipes.
+    P : array
+        Multipoles.
+    n_p : int
+        Total numper of pipes.
+    J : int
+        Number of multipoles per pipe to evaluate the thermal resistances.
+        J=1 or J=2 usually gives sufficient accuracy. J=0 corresponds to the
+        line source approximation.
+    r_b : float
+        Borehole radius (in meters).
+    r_p : float or array
+        Outer radius of the pipes (in meters).
+    z : array
+        Array of pipe coordinates in complex notation (x + 1.j*y). 
+    pikg : float
+        Inverse of 2*pi times the grout thermal conductivity, 1.0/(2.0*pi*k_g).
+    sigma : array
+        Dimensionless parameter for the ground and grout thermal
+        conductivities, (k_g - k_s)/(k_g + k_s).
+
+    Returns
+    -------
+    F : array
+        Matrix F_mk from Claesson and Hellstrom (2011), EQ. 34.
+
     """
     F = np.zeros((n_p, J), dtype=np.cfloat)
     for m in range(n_p):
@@ -1959,7 +2177,7 @@ def _F_mk(Q_p, P, n_p, J, r_b, r_p, z, pikg, sigma):
                                 *r_p[n]**(j+1)*(-r_p[m])**(k+1) \
                                 /(z[m] - z[n])**(j+k+2)
                     # Fourth term
-                    j_pend = np.min((k, j)) + 1
+                    j_pend = np.min((k, j)) + 2
                     for jp in range(j_pend):
                         fmk += sigma*np.conj(P[n,j])*binom(j+1, jp) \
                                 *binom(j+k-jp+1, j)*r_p[n]**(j+1) \
